﻿// Copyright 2014 Yves Desgraupes
// This file is part of BaseReader.
// 
// BaseReader is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// BaseReader is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with BaseReader.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.ComponentModel;
using System.Collections.ObjectModel;
using System.Windows.Input;
using System.Windows.Media;
using System.IO;
using Microsoft.Win32;

using BaseReaderPluginInterfaces;
using BaseReader.Properties;
using BaseReader.View;
using BaseReader.View.Controls;
using BaseReader.Managers;
using BaseReader.Tools;

namespace BaseReader
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow : Window, INotifyPropertyChanged
    {
#region INotifyPropertyChanged

        public event PropertyChangedEventHandler PropertyChanged;
        private void NotifyChange(String _PropertyName)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(_PropertyName));
        }

#endregion

#region Fields

        private TabItem m_lastTabItemTarget = null;
        private List<IBaseReaderItem> m_OpenedReaders = new List<IBaseReaderItem>();

        private String m_StatusDescription = String.Empty;
        
#endregion

#region Properties

        public String StatusDescription
        {
            get { return m_StatusDescription; }
            set { m_StatusDescription = value; NotifyChange("StatusDescription"); }
        }
        
#endregion

#region Methods
#region CDtors

        public MainWindow()
        {
            if (Settings.Default.HISTORY_ENTRIES == null)
                Settings.Default.HISTORY_ENTRIES = new System.Collections.Specialized.StringCollection();
            if (Settings.Default.OPENED_FILES == null)
                Settings.Default.OPENED_FILES = new System.Collections.Specialized.StringCollection();

            InitializeComponent();
            UpdateHistory();
            RestoreOpenedFiles();

            StatusDescription = String.Format("{0} plugin{1} loaded", PluginManager.Plugins.Count, (PluginManager.Plugins.Count > 1 ? "s" : ""));
        }

#endregion

#region Private

        private List<IBaseReaderItem> GetReaderItems(String[] _Files, out String _Errors, bool _SkipMissingFiles)
        {
            _Errors = String.Empty;
            List<IBaseReaderItem> newTabContentList = new List<IBaseReaderItem>();
            foreach (String filePath in _Files)
            {
                if (_SkipMissingFiles && File.Exists(filePath) == false)
                    continue;

                List<IBaseReaderPlugin> availablePlugins = PluginManager.MatchPlugins(filePath);
                List<IBaseReaderPlugin> pluginsToUse = null;
                if (availablePlugins.Count > 1)
                {
                    pluginsToUse = PluginSelector.ShowDialog(this,
                                                             availablePlugins,
                                                             String.Format("There is {0} plugins able to read the file \"{1}\".\r\n" +
                                                                           "Choose plugin(s) that will open it:",
                                                                           availablePlugins.Count, Path.GetFileName(filePath)),
                                                             "Available plugins");
                    // If the dialog return the same list (reference), the user canceled
                    if (pluginsToUse == availablePlugins || pluginsToUse == null)
                        continue;
                    availablePlugins = pluginsToUse;
                }
                List<String> loadingErrors;
                List<IBaseReaderItem> newItems = PluginManager.OpenFile(filePath, out loadingErrors, availablePlugins);
                if (newItems != null && newItems.Count != 0)
                    newTabContentList.AddRange(newItems);

                foreach (String err in loadingErrors)
                {
                    _Errors += err + "\r\n";
                }
            }

            return newTabContentList;
        }

        private int GenerateReaderControls(List<IBaseReaderItem> _ReaderItems)
        {
            int itemToFocus = -1;
            foreach (IBaseReaderItem elem in _ReaderItems)
            {
                AddNewHistory(elem.FilePath);
                IBaseReaderItem existingItem = m_OpenedReaders.Find(o => { return o.Equals(elem); });
                if (existingItem == null)
                {
                    m_OpenedReaders.Add(elem);
                    ClosableTabItem cti = new ClosableTabItem()
                    {
                        Title = Path.GetFileName(elem.FilePath),
                        Content = elem.ViewItem,
                        DataContext = elem
                    };
                    cti.OnClosing += PreviewTabItem_Closing;
                    cti.Close += TabItem_Close;

                    MainTabControl.Items.Add(cti);
                    itemToFocus = MainTabControl.Items.Count - 1;
                }
                else
                {
                    int idx = FindReaderIndexInTab(existingItem);
                    itemToFocus = (idx == -1 ? itemToFocus : idx);
                }
            }

            if (_ReaderItems.Count > 0)
            {
                UpdateHistory();
                StatusDescription = String.Format("{0} document{1} opened", _ReaderItems.Count, (_ReaderItems.Count > 1 ? "s" : ""));
            }
            else
                StatusDescription = "No document opened";

            return itemToFocus;
        }

        private void OpenFiles(String[] _Files, bool _SkipMissingFiles = false)
        {
            String errors;
            List<IBaseReaderItem> newTabContentList = GetReaderItems(_Files, out errors, _SkipMissingFiles);

            int itemToFocus = GenerateReaderControls(newTabContentList);

            if (!String.IsNullOrEmpty(errors))
            {
                MessageBox.Show(this, errors, "Unable to load the following file(s):", MessageBoxButton.OK, MessageBoxImage.Warning);
            }

            // Focus after message box to see the selection change (user confort)
            if (itemToFocus != -1)
            {
                MainTabControl.SelectedIndex = itemToFocus;
            }
        }

        private int FindReaderIndexInTab(IBaseReaderItem _ItemToFind)
        {
            if (_ItemToFind != null)
            {
                foreach (TabItem item in MainTabControl.Items)
                {
                    if (item != null && item.DataContext != null && item.DataContext is IBaseReaderItem)
                    {
                        if ((item.DataContext as IBaseReaderItem).Equals(_ItemToFind))
                            return MainTabControl.Items.IndexOf(item);
                    }
                }
            }
            return -1;
        }

        private void RestoreTabItem()
        {
            if (m_lastTabItemTarget == null)
                return;

            m_lastTabItemTarget.Margin = new Thickness(0, 0, 0, 0);
            m_lastTabItemTarget = null;
        }

        private void RestoreOpenedFiles()
        {
            if (Settings.Default.OPENED_FILES != null && Settings.Default.OPENED_FILES.Count > 0)
            {
                String[] files = new String[Settings.Default.OPENED_FILES.Count];
                Settings.Default.OPENED_FILES.CopyTo(files, 0);
                OpenFiles(files);
            }
        }

        private void UpdateHistory()
        {
            int sepIdx = FileMenuItem.Items.IndexOf(HistorySeparator);
            // Clear previous list
            for (int i = 0; i < FileMenuItem.Items.Count; ++i)
            {
                MenuItem mi = FileMenuItem.Items[i] as MenuItem;
                if (mi != null)
                {
                    if (mi.Tag is String && (mi.Tag as String) == "History")
                    {
                        FileMenuItem.Items.RemoveAt(i);
                        i--;
                    }
                }
            }

            if (sepIdx != -1)
            {
                // Order from younger to oldest
                for (int i = 0; i < Settings.Default.HISTORY_ENTRIES.Count; ++i)
                {
                    int pos = Settings.Default.HISTORY_ENTRIES.Count - i;
                    MenuItem mi = new MenuItem()
                    {
                        Header = String.Format("{0}: {1}", pos, Path.GetFileName(Settings.Default.HISTORY_ENTRIES[i])),
                        Tag = "History",
                        ToolTip = Settings.Default.HISTORY_ENTRIES[i],
                        DataContext = Settings.Default.HISTORY_ENTRIES[i]
                    };
                    mi.Click += MenuItem_HistoryOpen_Click;
                    FileMenuItem.Items.Insert(sepIdx + 1, mi);
                }
            }
        }

        private void AddNewHistory(String _NewEntry)
        {
            FileInfo fi = new FileInfo(_NewEntry);
            for (int i = 0; i < Settings.Default.HISTORY_ENTRIES.Count; ++i)
            {
                FileInfo oldFI = new FileInfo(Settings.Default.HISTORY_ENTRIES[i]);
                // Already registered, remove it, let the Add below append it
                if (oldFI.FullName.Equals(fi.FullName, StringComparison.OrdinalIgnoreCase))
                {
                    Settings.Default.HISTORY_ENTRIES.RemoveAt(i);
                    break;
                }
            }

            while (Settings.Default.HISTORY_ENTRIES.Count >= Settings.Default.HISTORY_MAX_ENTRIES)
                Settings.Default.HISTORY_ENTRIES.RemoveAt(0);

            Settings.Default.HISTORY_ENTRIES.Add(_NewEntry);
            Settings.Default.Save();
        }

        private void RestoreDnDState()
        {
            foreach (ClosableTabItem item in MainTabControl.Items)
            {
                item.HeaderGrabbed = false;
            }
            RestoreTabItem();
        }

        private void BroadcastToPlugins(System.Reflection.MethodInfo _Broadcaster, IBaseReaderItem _Item)
        {
            if (_Broadcaster == null)
                return;
            foreach (IBaseReaderPlugin plugin in PluginManager.Plugins)
            {
                if (plugin != null)
                {
                    _Broadcaster.Invoke(plugin, new object[] { _Item });
                }
            }
        }

#endregion

#region UI Events

        private void MenuItem_Open_Click(object sender, RoutedEventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Filter = "All Supported File |*.*";
            dlg.Multiselect = true;
            dlg.CheckFileExists = true;
            dlg.CheckPathExists = true;

            Nullable<bool> result = dlg.ShowDialog();
            if (result == true)
            {
                OpenFiles(dlg.FileNames);
            }
        }

        private void MenuItem_Exit_Click(object sender, RoutedEventArgs e)
        {
            this.Close();
        }

        private void MenuItem_FileInfo_Click(object sender, RoutedEventArgs e)
        {

        }

        private void MenuItem_HistoryOpen_Click(object sender, RoutedEventArgs e)
        {
            MenuItem mi = sender as MenuItem;
            if (mi != null && mi.DataContext is String)
            {
                String path = mi.DataContext as String;
                OpenFiles(new String[] { path });
            }
        }

        private void PreviewTabItem_Closing(object sender, CancelEventArgs e)
        {
            if (sender is ClosableTabItem)
            {
                ClosableTabItem ti = sender as ClosableTabItem;
                if (ti.IsModified)
                {
                    if (MessageBox.Show(this,
                                        "The document \"" + ti.Title + "\" have modification.\r\n" +
                                        "Do you really want to close this document?",
                                        "Closing modified document",
                                        MessageBoxButton.OKCancel,
                                        MessageBoxImage.Question) == MessageBoxResult.Cancel)
                    {
                        e.Cancel = true;
                    }
                }
            }
        }

        private void TabItem_Close(ClosableTabItem _Sender)
        {
            bool found = false;
            foreach (ClosableTabItem item in MainTabControl.Items)
            {
                if (item == _Sender)
                {
                    found = true;
                    break;
                }
            }

            if (found)
            {
                MainTabControl.Items.Remove(_Sender);
                if (_Sender.DataContext is IBaseReaderItem)
                {
                    IBaseReaderItem reader = (_Sender.DataContext as IBaseReaderItem);
                    BroadcastToPlugins(typeof(IBaseReaderPlugin).GetMethod("OnFileClosed"), reader);
                    reader.Dispose();
                    m_OpenedReaders.Remove(reader);
                }
            }
        }

        private void MenuItem_OpenContainingFolder_Click(object sender, RoutedEventArgs e)
        {
            ClosableTabItem cti = MainTabControl.SelectedItem as ClosableTabItem;
            if (cti != null)
            {
                IBaseReaderItem item = cti.DataContext as IBaseReaderItem;
                if (item != null)
                {
                    ToolIO.OpenDirectoryAndSelectFile(item.FilePath);
                }
            }
        }

        private void Window_Closed(object sender, EventArgs e)
        {
            Settings.Default.OPENED_FILES.Clear();
            foreach (IBaseReaderItem item in m_OpenedReaders)
            {
                Settings.Default.OPENED_FILES.Add(item.FilePath);
            }
            Settings.Default.MAINWINDOW_WIDTH = this.Width;
            Settings.Default.MAINWINDOW_HEIGHT = this.Height;
            Settings.Default.Save();
        }
        
#region DnD

        private void Window_DragOver(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                e.Effects = DragDropEffects.Copy;
            }
            else
            {
                e.Effects = DragDropEffects.None;
            }
        }

        private void Window_Drop(object sender, DragEventArgs e)
        {
            if (e.Data.GetDataPresent(DataFormats.FileDrop))
            {
                e.Effects = DragDropEffects.Copy;
                String[] filesList = e.Data.GetData(DataFormats.FileDrop) as String[];
                OpenFiles(filesList);
            }
            else
            {
                e.Effects = DragDropEffects.None;
            }

            RestoreDnDState();
        }

        private void TabItem_PreviewMouseMove(object sender, MouseEventArgs e)
        {
            ClosableTabItem tabItem = null;

            var trg = e.Source as DependencyObject;
            if (trg != null)
                tabItem = trg.TryFindParent<ClosableTabItem>();

            if (tabItem != null && tabItem.HeaderGrabbed && Mouse.PrimaryDevice.LeftButton == MouseButtonState.Pressed)
            {
                DragDrop.DoDragDrop(tabItem, tabItem, DragDropEffects.All);
            }
        }

        private void TabItem_DragOver(object sender, DragEventArgs e)
        {
            ClosableTabItem tabItemTarget = null;
            ClosableTabItem tabItemSource = e.Data.GetData(typeof(ClosableTabItem)) as ClosableTabItem;

            var trg = e.Source as DependencyObject;
            if (trg != null)
                tabItemTarget = trg.TryFindParent<ClosableTabItem>();

            if (tabItemTarget != tabItemSource && tabItemSource != null && tabItemTarget != null)
            {
                if (m_lastTabItemTarget != tabItemTarget)
                    RestoreTabItem();

                tabItemTarget.Margin = new Thickness(4, 0, 0, 0);
                m_lastTabItemTarget = tabItemTarget;
            }
            else
                RestoreTabItem();
        }

        private void TabItem_Drop(object sender, DragEventArgs e)
        {
            ClosableTabItem tabItemTarget = null;
            ClosableTabItem tabItemSource = e.Data.GetData(typeof(ClosableTabItem)) as ClosableTabItem;
            var trg = e.Source as DependencyObject;
            if (trg != null)
                tabItemTarget = trg.TryFindParent<ClosableTabItem>();

            if (tabItemTarget != tabItemSource && tabItemSource != null && tabItemTarget != null)
            {
                tabItemSource.HeaderGrabbed = false;
                tabItemTarget.HeaderGrabbed = false;

                var tabControl = tabItemTarget.Parent as TabControl;
                int sourceIndex = tabControl.Items.IndexOf(tabItemSource);
                int targetIndex = tabControl.Items.IndexOf(tabItemTarget);

                tabControl.Items.Remove(tabItemSource);
                tabControl.Items.Insert(targetIndex, tabItemSource);

                tabControl.Items.Remove(tabItemTarget);
                tabControl.Items.Insert(sourceIndex, tabItemTarget);
            }
            RestoreTabItem();
        }

        private void Window_PreviewMouseMove(object sender, MouseEventArgs e)
        {
            if (Mouse.PrimaryDevice.LeftButton != MouseButtonState.Pressed)
            {
                RestoreDnDState();
            }
        }

        private void MainTabControl_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if (sender == MainTabControl &&
                e.Source == MainTabControl &&
                MainTabControl.SelectedItem is FrameworkElement)
            {
                FrameworkElement fe = MainTabControl.SelectedItem as FrameworkElement;
                if (fe.DataContext is IBaseReaderItem)
                {
                    IBaseReaderItem item = fe.DataContext as IBaseReaderItem;
                    BroadcastToPlugins(typeof(IBaseReaderPlugin).GetMethod("OnFileSelected"), item);

                }
            }
        }

#endregion
#endregion
#endregion
    }
}
